# (C) 2024, Tom Eulenfeld, MIT license

import argparse
import contextlib
from copy import deepcopy
from importlib.resources import files
import logging
import logging.config
from pathlib import Path
import os
import subprocess
import sys
import tempfile
import time
import tomllib
from warnings import warn

from sugar import read
from anchorna.core import combine, cutout, find_my_anchors
from anchorna.io import jalview_features, read_anchors, load_selected_anchors
from anchorna.util import _apply_mode, Options

@contextlib.contextmanager
def _changedir(path):
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


EXAMPLE_TOML_CONFIG = """### Example configuration for anchorna go command in TOML format
# The fname configuration value can be used by some other anchorna commands.

fname = "pesti_example.gff"
w = 5  # word length
refid = "KC533775"
maxshift = 100
scoring = "blosum62"
thr_score = 19  # add fluke only if above this threshold
thr_quota_score = 22  # count number of flukes with a score >= thr_quota_score
thr_quota = 1  # discard anchor if portion of counted flukes is smaller than quota (1=100%)

removed_anchors_path = "removed_anchors_{}.gff"  # turn off with "none", alternatively delete this line
logfile = "anchorna.log"  # turn off with "none", alternatively delete this line
"""


LOGGING_DEFAULT_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'capture_warnings': True,
    'formatters': {
        'file': {
            'format': ('%(asctime)s  %(levelname)s:%(name)s:%(message)s'),
            'datefmt': '%Y-%m-%d %H:%M:%S'
        },
        'console': {
            'format': '%(levelname)s:%(message)s'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'console',
            'level': 'DEBUG',
        },
        'file': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'level': 'DEBUG',
            'filename': None,
        },
    },
    'loggers': {
        'anchorna': {
            'handlers': ['console', 'file'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'py.warnings': {
            'handlers': ['console', 'file'],
            'level': 'DEBUG',
            'propagate': False,
        }
    }
}


def _configure_logging(loggingc, logfile=None):
    if loggingc is None:
        loggingc = deepcopy(LOGGING_DEFAULT_CONFIG)
        if logfile is None or logfile.lower() in ('none', 'null'):
            del loggingc['handlers']['file']
            loggingc['loggers']['anchorna']['handlers'] = ['console']
            loggingc['loggers']['py.warnings']['handlers'] = ['console']
        else:
            loggingc['handlers']['file']['filename'] = logfile
    logging.config.dictConfig(loggingc)
    logging.captureWarnings(loggingc.get('capture_warnings', False))


def _start_ipy(obj):
    from IPython import start_ipython
    print('Anchors loaded into anchors variable.')
    start_ipython(argv=[], user_ns={'anchors': obj}, display_banner=False)
    print('Bye')


def _tutorial_seqs(subset=False):
    seqs = read(files('anchorna.tests.data').joinpath('pesti56.gff.zip'))
    if subset:
        seqs = seqs[18:28]
        start, stop = zip(*[seq.fts.get('cds').loc_range for seq in seqs])
        start = max(start)
        stop = min(stop)
        stop = start + (stop-start) // 3 * 3 + 6
        for i in range(len(seqs)):
            seqs[i] = seqs[i][:start + 999] + seqs[i][stop:]
            seqs[i].fts[0].loc.stop -= stop-start - 999
    return seqs


def _cmd_create(conf, tutorial=False, tutorial_subset=False):
    if tutorial and tutorial_subset:
        raise ValueError('Only one of tutorial or tutorial_subset are allowed')
    with open(conf, 'w') as f:
        f.write(EXAMPLE_TOML_CONFIG)
    if tutorial or tutorial_subset:
        seqs = _tutorial_seqs(subset=tutorial_subset)
        path = os.path.dirname(conf)
        seqs.write(os.path.join(path, 'pesti_example.gff'))

def _cmd_go(fname, fname_anchor, pbar=True, remove=True, continue_with=None,
            removed_anchors_path=None, logconf=None, logfile=None, **kw):
    _configure_logging(logconf, logfile)
    log = logging.getLogger('anchorna')
    options = Options(**kw)
    log.info(f'anchorna go is called with arguments {fname=}, {fname_anchor=}, '
             f'{remove=}, {continue_with=}, {removed_anchors_path=}')
    log.info(f'{options=}')
    seqs = read(fname)
    if continue_with is not None:
        continue_with = read_anchors(continue_with)
    anchors, removed_anchors = find_my_anchors(
        seqs, options, remove=remove, continue_with=continue_with, pbar=pbar)
    rap = removed_anchors_path
    if remove:
        if rap is None or rap.lower() in ('none', 'null'):
            log.debug('Do not save removed anchors')
        else:
            rap = rap.format(time.strftime('%Y_%m_%d-%H_%M_%S'))
            log.info(f'Save removed anchors in file {rap}')
            removed_anchors.write(rap)
    else:
        assert removed_anchors is None
    log.info(f'Write anchors to file {fname_anchor}')
    anchors.write(fname_anchor)

def _cmd_print(fname_anchor, verbose=False, mode='aa'):
    anchors = load_selected_anchors(fname_anchor)
    try:
        print(anchors.tostr(verbose=verbose, mode=mode))
    except BrokenPipeError:
        pass

def _cmd_load(fname_anchor):
    _start_ipy(read_anchors(fname_anchor))

def _cmd_export(fname_anchor, out, mode='aa'):
    assert mode in ('seq', 'cds', 'aa')
    anchors = load_selected_anchors(fname_anchor)
    outstr = jalview_features(anchors, mode=mode)
    if out is None:
        print(outstr)
    else:
        with open(out, 'w') as f:
            f.write(outstr)

def _cmd_view(fname_anchor, fname, mode='aa', align=None):
    assert mode in ('seq', 'cds', 'aa')
    with tempfile.TemporaryDirectory(prefix='anchorna') as tmpdir:
        fname_export = Path(tmpdir) / 'jalview_features.txt'
        fnameseq = Path(tmpdir) / 'aa_or_seq_or_cds.fasta'
        _cmd_export(fname_anchor, fname_export, mode=mode)
        seqs = read(fname)
        if mode != 'seq':
            anchors = read_anchors(fname_anchor)
            offsets = {f.seqid: f.offset for anchor in anchors for f in (anchor.data + anchor.missing)}
            for seq in seqs:
                seq.data = seq.data[offsets[seq.id]:]
            if mode == 'aa':
                seqs = seqs.translate(check_start=False)
        if align:
            anchors = read_anchors(fname_anchor)
            anchor = anchors[int(align.lower().removeprefix('a'))]
            anchor.data += anchor.missing
            anchor.missing = []
            start = max(_apply_mode(fluke.start, fluke.offset, mode=mode) for fluke in anchor)
            for seq in seqs:
                fluke = anchor.d[seq.id]
                seq.data = '-' * (start - _apply_mode(fluke.start, fluke.offset, mode=mode)) + seq.data
        seqs.write(fnameseq)
        subprocess.run(f'jalview {fnameseq} --features {fname_export}'.split())

def _cmd_combine(fname_anchor, out):
    lot_of_anchors = [load_selected_anchors(fn) for fn in fname_anchor]
    anchors = combine(lot_of_anchors)
    anchors.write(out)

def _cmd_cutout(fname, fname_anchor, pos1, pos2, out, fmt, mode='seq'):
    assert mode in ('seq', 'cds', 'aa')
    seqs = read(fname)
    anchors = load_selected_anchors(fname_anchor)
    seqs2 = cutout(seqs, anchors, pos1, pos2, mode=mode)
    if out is None:
        print(seqs2.tofmtstr(fmt or 'fasta'))
    else:
        seqs2.write(out, fmt)


def run(command, conf=None, pdb=False, **args):
    if command == 'create':
        return _cmd_create(conf, **args)
    if conf:
        try:
            with open(conf, 'rb') as f:
                conf = tomllib.load(f)
        except ValueError as ex:
            sys.exit('Error while parsing the configuration: %s' % ex)
        except IOError as ex:
            if command == 'go':
                warn(ex)
            conf = {}
    else:
        conf = {}
    if command in ('cutout', 'view'):
        # ignore all config settings except fname
        conf = {'fname': conf['fname']} if 'fname' in conf else {}
    # Populate args with conf, but prefer args
    conf.update(args)
    args = conf
    if pdb:
        import traceback
        import pdb
        def info(type, value, tb):
            traceback.print_exception(type, value, tb)
            print()
            pdb.pm()
        sys.excepthook = info
    try:
        globals()[f'_cmd_{command}'](**args)
    except KeyError:
        raise ValueError(f'Unknown command: {command}')


def run_cmdline(cmd_args=None):
    """Main entry point from the command line"""
    from anchorna import __version__
    msg = 'anchoRNA: Find anchors in short sequences of RNA/DNA'
    epilog = ('To get help on a subcommand run: anchorna command -h. '
              'Each time you provide a anchor file, you may load only '
              'selected anchors from the file using a dedicated syntax, '
              'fname_anchor|select or fname_anchor|select|remove, e.g. '
              'fname_anchor|4:10,12 only loads anchors 4:10, 12 (python indexing) - '
              'fname_anchor||a3 will remove the anchor a3 and fname_anchor|0:10|3 will '
              'select the first 10 anchors and remove a3. '
              'Numbers may be prepended with a single "a". '
              'The character - can be used to load a anchor file from stdin.')
    parser = argparse.ArgumentParser(description=msg, epilog=epilog)
    version = '%(prog)s ' + __version__
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument('--pdb', action='store_true', help='Start the debugger upon exception')

    sub = parser.add_subparsers(title='commands', dest='command')
    sub.required = True

    p_create = sub.add_parser('create', help=(msg:='create example configuration'), description=msg)
    p_go = sub.add_parser('go', help=(msg:='find anchors and write them into anchor file'), description=msg)
    p_print = sub.add_parser('print', help=(msg:='print contents of anchor file'), description=msg)
    p_load = sub.add_parser('load', help=(msg:='load anchors into IPython session'), description=msg)
    p_export = sub.add_parser('export', help=(msg:='export anchors into feature file to view them in Jalview'), description=msg)
    p_view = sub.add_parser('view', help=(msg:='view anchors in JalView (i.e. call export and JalView)'), description=msg)
    msg = 'combine (and select or remove) anchors from one file or different files'
    msg2 = ('An expression consists of a file name fname_anchor or fname_anchor|selection to select anchors or fname_anchor||expression to remove anchors, or fname|selection|expression. '
            'Example call: anchorna combine f1.gff "f2.gff|a3:a5,a7" "f3.gff|a10" --out f5.gff, see also anchorna -h')
    p_combine = sub.add_parser('combine', help=msg, description=msg, epilog=msg2)
    msg = 'cut out sequences between anchors and export them to new sequences file'
    msg2 = ('This command serves two purposes: '
            '1. Cutout sequences to put them again into the go command and '
            'find more anchors with updated options. '
            'Please use a file format which preserves the meta.offset attributes, e.g. sjson; use mode seq.'
            'Combine the found anchors with the anchorna combine command. '
            '2. Cutout sequences to use them in external tool. '
            'The cutout command expects two positions in the sequence. '
            'Each possition has 3 parts ABC where B and C are optional. '
            'Part A: Is a number or number prepended with letter a to specify the anchor number, '
            'use special words "start" and "end" for start or end of sequence, '
            'use special words "ATG" and "*" for start or stop codon of sequence (only allowed in mode "seq") '
            'Part B: One of the characters <, >, ^, for start, end or middle of word (anchor) specified in A, '
            'default is ^ (middle), must be ommitted for A=start or A=end. '
            'Part C: Additional character offset in the form +X or -X.'
            'Examples: anchorna cutout anchors.gff "a11<" "a12>+10", anchorna cutout anchors.gff a10 "*>"'
            )
    p_cutout = sub.add_parser('cutout', help=msg, description=msg, epilog=msg2)


    for p in (p_create, p_go, p_cutout, p_view):
        p.add_argument('-c', '--conf', default='anchorna.conf', help='configuration file to use (default: anchorna.conf)')
    p_create.add_argument('--tutorial', action='store_true', help='copy GFF sequences file for tutorial')
    p_create.add_argument('--tutorial-subset', action='store_true', help=argparse.SUPPRESS)  # for testing purposes
    p_go.add_argument('fname_anchor', help='anchor file name (GFF output)')
    p_go.add_argument('--no-pbar', help='do not show progress bar', action='store_false', dest='pbar', default=argparse.SUPPRESS)
    p_go.add_argument('--no-remove', help='do not remove contradicting anchors', action='store_false', dest='remove', default=argparse.SUPPRESS)
    p_go.add_argument('--continue-with', help='only remove contradicting anchors from passed anchor list', default=argparse.SUPPRESS)

    g = p_go.add_argument_group('optional arguments', description='Use these flags to overwrite values in the config file.')
    features = [(int, ('w', 'maxshift')),
                (str, ('fname', 'refid', 'scoring', 'logfile', 'removed-anchors-path')),
                (float, ('thr-score', 'thr-quota-score', 'thr-quota'))]
    for type_, fs in features:
        for f in fs:
            g.add_argument('--' + f, default=argparse.SUPPRESS, type=type_)
    for p in (p_cutout, p_view):
        g = p.add_argument_group('optional arguments', description='Use these flags to overwrite values in the config file.')
        g.add_argument('--fname', default=argparse.SUPPRESS)


    p_export.add_argument('fname_anchor', help='anchor file name')
    p_export.add_argument('-o', '--out', help='output file name (by default prints to stdout)')
    p_view.add_argument('fname_anchor', help='anchor file name')
    p_view.add_argument('--align', help='align sequences at given anchor')
    p_combine.add_argument('fname_anchor', nargs='+', help='anchor file name')
    p_combine.add_argument('-o', '--out', help='output file name (by default prints to stdout)')

    p_cutout.add_argument('fname_anchor', help='anchor file name')
    p_cutout.add_argument('pos1', help='left position')
    p_cutout.add_argument('pos2', help='right position')

    p_cutout.add_argument('-o', '--out', help='output file name (by default prints fasta to stdout)')
    p_cutout.add_argument('-f', '--fmt', help='format of file (default: autodetect)')

    p_print.add_argument('fname_anchor', help='anchor file name')
    p_print.add_argument('-v', '--verbose', help=msg, action='store_true')
    p_load.add_argument('fname_anchor', help='anchor file name')

    for p in (p_print, p_export, p_view, p_cutout):
        choices = ['seq', 'cds', 'aa']
        default = 'seq' if p == p_cutout else 'aa'
        msg = ('choose mode, seq: relative to original sequence, '
               'cds: accounting for offset (usually coding sequence), aa: translated cds '
               f'(default: {default})')
        p.add_argument('-m', '--mode', default=default, choices=choices, help=msg)

    # Get command line arguments and start run function
    args = parser.parse_args(cmd_args)
    run(**vars(args))

if __name__ == '__main__':
    run_cmdline()



# (C) 2024, Tom Eulenfeld, MIT license
import json
import logging
import sys
from warnings import warn

import matplotlib.colors as mcolors
from matplotlib.colors import to_hex, to_rgb
from sugar import read_fts

from anchorna.util import _apply_mode, fts2anchors, Anchor, AnchorList, Fluke, Options


log = logging.getLogger('anchorna')


def write_anchors(anchors, fname):
    """
    Write anchors to GFF file

    Offsets are stored as comments.
    """
    from anchorna import __version__
    fts = anchors.convert2fts()
    offsets = {ft.seqid: ft.meta._gff.pop('offset') for ft in fts}
    offsets_header = ''.join(f'#offset {seqid} {offset}\n' for seqid, offset in offsets.items())
    header = f'#AnchoRNA anchor file\n#written with AnchoRNA v{__version__}\n' + offsets_header
    if fname is None:
        try:
            print(fts.tofmtstr('gff', header=header))
        except BrokenPipeError:
            pass
    else:
        fts.write(fname, 'gff', header=header)


def read_anchors(fname):
    """
    Read anchors from GFF file

    Offsets are restored from comments.
    """
    comments = []
    fts = read_fts(fname, 'gff', comments=comments)
    offsets = {}
    for i, line in enumerate(comments):
        if i == 0:
            if not line.startswith('##gff-version 3'):
                warn(f'{fname} not a valid GFF file')
        elif i == 1:
            if not line.startswith('#AnchoRNA'):
                warn(f'{fname} not a valid anchor file')
        elif line.startswith('#offset'):
            seqid, offset = line.split()[1:]
            offsets[seqid] = int(offset)
        elif line.startswith('#'):
            continue
        else:
            break
    for ft in fts:
        ft.meta._gff.offset = offsets[ft.seqid]
    return fts2anchors(fts)


def _parse_selection(anchors, selection):
    anchors2 = anchors[:0]
    for part in selection.split(','):
        if ':' in part:
            i, j = part.split(':')
            i = int(i.strip().removeprefix('a'))
            j = int(j.strip().removeprefix('a'))
        else:
            i = int(part.strip().removeprefix('a'))
            j = i + 1
        anchors2.extend(anchors[i:j])
    return anchors2


def load_selected_anchors(fname):
    """
    Read anchors and select or remove some of them
    """
    if '|' not in fname:
        return read_anchors(fname)
    fname, selection = fname.split('|', 1)
    anchors = read_anchors(fname.strip())
    selection = selection.lower()
    if '|' not in selection:
        return _parse_selection(anchors, selection)
    selection, rem = selection.split('|')
    anchors_remove = set(_parse_selection(anchors, rem))
    anchors = _parse_selection(anchors, selection) if selection.strip() else anchors
    anchors.data = [a for a in anchors if a not in anchors_remove]
    return anchors


class _AnchorJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (Anchor, AnchorList, Fluke, Options)):
            obj = {k: v for k, v in o.__dict__.items()}
            obj['_cls'] = type(o).__name__
            return obj
        else:
            # Let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, o)


class _ConfigJSONDecoder(json.JSONDecoder):
    """Decode JSON config with comments stripped"""
    def decode(self, s):
        s = '\n'.join(l.split('#', 1)[0] for l in s.split('\n'))
        return super(_ConfigJSONDecoder, self).decode(s)


def _json_hook(d):
    if cls := d.pop('_cls', None):
        return globals()[cls](**d)
    else:
        return d


def load_json(fname):
    """
    Load anchors from JSON file
    """

    if fname == '-':
        return json.loads(sys.stdin.read(), object_hook=_json_hook)
    else:
        with open(fname) as f:
            return json.load(f, object_hook=_json_hook)


def write_json(obj, fname):
    """
    Write anchors to JSON file
    """
    if fname is None:
        print(json.dumps(obj, cls=_AnchorJSONEncoder))
    else:
        with open(fname, 'w') as f:
            json.dump(obj, f, cls=_AnchorJSONEncoder)


def _make_rgb_transparent(rgb, bg_rgb, alpha):
    return [alpha * c1 + (1 - alpha) * c2
            for (c1, c2) in zip(to_rgb(rgb), to_rgb(bg_rgb))]


cols = list(mcolors.TABLEAU_COLORS.values())
to_hex(cols[0], 0.3)

def jalview_features(anchors, mode='aa'):
    """
    Write anchors to Jalview feature file
    """
    assert mode in ('seq', 'cds', 'aa')
    anchors = sorted(anchors, key=lambda a: a.ref.start)
    content = []
    header = []
    #for winlen, *_, aaref, r in anchors:
    for k, a in enumerate(anchors):
        #for seqid, (score, i, j, _) in r.items():
        for j, f in enumerate(a):
            # anchor2	C_CSFV_KC533775	-1	130	150	anchorsim	1.0
            # anhcor100	D_BDV_NC_003679	-1	10	20	anchorxx
            wlen = a.ref.len
            score_ = f.score / a.maxscore
            c = to_hex(_make_rgb_transparent(cols[k % len(cols)], 'white', score_)).strip('#')

            al = f'anchor{k}_s{f.score}'
            miss = f'misses:{len(a.missing)}'
            header.append(
                f'{al}\t{c}\n'
                )
            i = _apply_mode(f.start, f.offset, mode)
            j = _apply_mode(f.stop, f.offset, mode)
            content.append(
                f'{a.ref.word[:5]} w{wlen} {miss}\t{f.seqid}\t-1\t{i+1}\t{j}\t{al}\n')
                 #f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{i+1}\t{i+wlen}\t{al}\n' if mode == 'default' else
                 #f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{i+o+1}\t{i+o+wlen}\t{al}\n' if mode == 'offset' else
                 #f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{3*i+1}\t{3*i+3*wlen}\t{al}\n' if mode == 'seq' else
                 #f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{3*i+o+1}\t{3*i+o+3*wlen}\t{al}\n' if mode == 'seqoffset' else
                 #None)
    header.append('\nSTARTFILTERS\nENDFILTERS\n\n')
    return ''.join(header) + ''.join(content)



# (C) 2024, Tom Eulenfeld, MIT license

import collections
import itertools
import logging
from statistics import median
from warnings import warn

from sugar import Attr
from sugar.data import submat


log = logging.getLogger('anchorna')


def corrscore(seq1, seq2, gap='-', sm=submat('blosum62')):
    return sum(sm[nt1][nt2] for nt1, nt2 in zip(seq1, seq2) if nt1 != gap and nt2 != gap)


class Options(Attr):
    pass


class Fluke(Attr):
    @property
    def len(self):
        return self.stop - self.start


def _apply_mode(i, o, mode, islen=False):
    if mode == 'aa':
        return i
    elif mode == 'cds' or mode == 'seq' and islen:
        return 3 * i
    elif mode == 'seq':
        return 3 * i + o
    raise ValueError('mode not allowed')


class Anchor(collections.UserList):
    def __init__(self, data=None, **kw):
        super().__init__([fluke for fluke in data if fluke.above_thres])
        for k, v in kw.items():
            setattr(self, k, v)
        self.missing = [fluke for fluke in data if not fluke.above_thres]

    def __str__(self):
        return self.tostr()

    @property
    def id(self):
        return f'A {self.ref.start}+{self.ref.len}'

    def __hash__(self):
        return hash((self.ref.start, self.ref.len, self.ref.offset))

    def todict(self):
        return {f.seqid: f for f in self}

    @property
    def d(self):
        return self.todict()

    def tostr(self, i='', verbose=False, mode='aa'):
        ind = _apply_mode(self.ref.start, self.ref.offset, mode=mode)
        len_ = _apply_mode(self.ref.len, self.ref.offset,
                           mode=mode, islen=True)
        out = f'A{i} {ind}+{len_}  minscore {self.minscore}  misses {len(self.missing)}  {self.ref.word}'
        if not verbose:
            return out
        flukes = [f'  F{j} {_apply_mode(f.start, f.offset, mode=mode)}  score {f.score}  {f.word}  {f.seqid}' + '  (below thresshold)' * (not f.above_thres)
                  for j, f in enumerate(self)]
        flukesm = [f'  -- {_apply_mode(f.start, f.offset, mode=mode)}  score {f.score}  {f.word}  {f.seqid}' + '  (below thresshold)' * (not f.above_thres)
                  for j, f in enumerate(self.missing)]
        return '\n'.join([out]+flukes + flukesm)

    @property
    def ref(self):
        return self.d[self.refid]

    @property
    def minscore(self):
        return min(f.score for f in self)

    @property
    def maxscore(self):
        return max(f.score for f in self)

    @property
    def medscore(self):
        return median(f.score for f in self)

    def overlaps_with(self, a2):
        def key(f): return '' if f.seqid == a1.refid else f.seqid
        a1 = self
        return (max(a1.ref.start, a2.ref.start) <= min(a1.ref.stop, a2.ref.stop) and
                {f.seqid for f in a1.missing} == {f.seqid for f in a2.missing} and
                {f.seqid for f in a1} == {f.seqid for f in a2} and
                all(f1.start - f2.start == a1.ref.start - a2.ref.start for f1, f2 in zip(sorted(a1, key=key), sorted(a2, key=key))))

    def join_with(self, a2):
        a1 = self
        if not a1.overlaps_with(a2):
            raise ValueError('Cannot join anchors which do not overlap')
        if a1.ref.start <= a2.ref.start and a1.ref.stop >= a2.ref.stop:
            # a2 is contained in a1
            return a1
        elif a1.ref.start >= a2.ref.start and a1.ref.stop <= a2.ref.stop:
            # a1 is contained in a2
            return a2
        else:
            # overlap
            def key(f): return not f.above_thres, '' if f.seqid == a1.refid else f.seqid
            flukes = []
            correctl = None
            for f1, f2 in zip(sorted(a1.data + a1.missing, key=key), sorted(a2.data + a2.missing, key=key)):
                assert f1.seqid == f2.seqid
                assert f1.above_thres == f2.above_thres
                assert f1.offset == f2.offset
                start = min(f1.start, f2.start)
                stop = max(f1.stop, f2.stop)
                l = max(f1.stop, f2.stop) - start
                if f1.above_thres:
                    if correctl is None:
                        correctl = l
                    assert correctl == l
                overlaplen = f1.len + f2.len - l
                if correctl != l:
                    # there is a gap between the flukes!
                    # that might only be the case for "missing" flukes below the thresshold
                    # adapt the word
                    assert not f1.above_thres and not f2.above_thres
                    l = correctl
                    stop = start + l
                    overlaplen = f1.len + f2.len - l
                # old way to calculate the score: estimation in the following way, we do not recalculate it
                #score = f1.score * (1 - overlaplen / 2 / f1.len) + f2.score * (1 - overlaplen / 2 / f2.len)
                nword = f1.word + f2.word[overlaplen:] if f1.start <= f2.start else f2.word + f1.word[overlaplen:]
                assert len(nword) == l
                fluke = Fluke(seqid=f1.seqid, score=None,
                              start=start, stop=stop, offset=f1.offset, word=nword,
                              above_thres=f1.above_thres or f2.above_thres)
                flukes.append(fluke)
            anchor = Anchor(flukes, refid=a1.refid)
            anchor._calculate_scores()
            return anchor


    def _calculate_scores(self):
        words = {fluke.word for fluke in self if fluke.above_thres}
        for fluke in self.data + self.missing:
            fluke.score = median(corrscore(fluke.word, w) for w in words)


    def contradicts(self, a2):
        a1 = self
        def key(f): return '' if f.seqid == a1.refid else f.seqid
        if a1.ref.start > a2.ref.start:
            a1, a2 = a2, a1
        return not all(f1.start <= f2.start for f1, f2 in zip(sorted(a1, key=key), sorted(a2, key=key)))


class AnchorList(collections.UserList):
    def __init__(self, data=None, **kw):
        super().__init__(data)
        for k, v in kw.items():
            setattr(self, k, v)

    def __str__(self):
        return self.tostr()

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('...')
        else:
            p.text(str(self))

    def tostr(self, verbose=False, mode='aa'):
        return '\n'.join(a.tostr(i=i, verbose=verbose, mode=mode) for i, a in enumerate(self))

    def merge_neighbor_anchors(self):
        while True:
            for a1, a2 in itertools.combinations(sorted(self, key=lambda a: a.minscore, reverse=True), 2):
                if a1.overlaps_with(a2):
                    break
            else:
                break
            # merge a1 and a2
            self.data.remove(a1)
            self.data.remove(a2)
            a = a1.join_with(a2)
            self.data.append(a)
        return self

    def remove_contradicting_anchors(self):
        removed_anchors = []
        a1s = sorted(self, key=lambda a: a.minscore, reverse=True)
        a2s = sorted(self, key=lambda a: a.minscore)
        while True:
            for a1, a2 in itertools.product(a1s, a2s):
                if a1 == a2 or a1.minscore < a2.minscore:
                    continue
                if a1.contradicts(a2):
                    break
            else:
                break
            # remove a2
            # remove one of a1, a2
            # if a1.minscore < a2.minscore:
            #     a1, a2 = a2, a1
            log.debug(f'Remove anchor {a2.ref.start}+{a2.ref.len} with min score {a2.minscore}, '
                      f'keep anchor {a1.ref.start}+{a1.ref.len} with min score {a1.minscore}')
            self.data.remove(a2)
            a2s.remove(a2)
            a1s.remove(a2)
            removed_anchors.append(a2)
        return AnchorList(removed_anchors)

    def convert2fts(self):
        return anchors2fts(self)

    def write(self, fname):
        from anchorna.io import write_anchors
        return write_anchors(self, fname)


def anchors2fts(anchors):
    """
    Convert anchors to sugar.FeatureList object

    Write some important attributes to Feature.meta._gff metadata.
    """
    from sugar import Feature, FeatureList
    fts = []
    for i, a in enumerate(anchors):
        def key(f): return '' if f.seqid == a.refid else f.seqid
        data = sorted(a.data + a.missing, key=key)
        for j, f in enumerate(data):  # fluke
            if a.refid == f.seqid:
                assert j == 0
                ftype = 'anchor'
                name = f'A{i}'
            else:
                assert j > 0
                ftype = 'fluke'
                name=f'A{i}_{f.seqid}'
            ft = Feature(ftype, start=f.start, stop=f.stop,
                         meta=dict(name=name, seqid=f.seqid, score=f.score,
                                   _gff=Attr(source='anchorna', word=f.word)))
            if hasattr(f, 'offset'):
                ft.meta._gff.offset = f.offset
            if not f.above_thres:
                ft.meta._gff.below_thres = 1
            fts.append(ft)
    return FeatureList(fts)


def fts2anchors(fts):
    """
    Convert sugar.FeatureList object to anchors

    Read some important attributes from Feature.meta._gff metadata.
    """
    anchors = []
    flukes = []
    refid = None
    for ft in fts:
        if ft.type == 'anchor':
            if len(flukes) > 0:
                anchors.append(Anchor(flukes, refid=refid))
                flukes = []
            refid = ft.seqid
            aname = ft.meta.name
        if ft.type in ('anchor', 'fluke'):
            if not ft.meta.name.startswith(aname):
                raise ValueError(f'Fluke with name {ft.meta.name} not part of anchor {aname}')
            start, stop = ft.loc_range
            fluke = Fluke(seqid=ft.seqid, score=ft.meta.score, start=start, stop=stop,
                          word=ft.meta._gff.word, above_thres=not hasattr(ft.meta._gff, 'below_thres'))
            if hasattr(ft.meta._gff, 'offset'):
                fluke.offset = ft.meta._gff.offset
            flukes.append(fluke)
        else:
            warn(f'Cannot convert feature of type {ft.type}')
    if len(flukes) > 0:
        anchors.append(Anchor(flukes, refid=refid))
    return AnchorList(anchors)


# def write_anchortxt(anchors, fname):
#     with open(fname, 'w') as f:
#         print('# winlen score index(aa) index(nt) aa nt', file=f)
#         for a in anchors:
#             print(*a[:-1], file=f)


# def __write_jalview_features_simple(anchors, fnameaa, fnament, fnamecds):
#     header = ('anchor\tlabel\n'
#               'anchorsim\tscore|bad9f7|105191|absolute|0.7|1\n\n'
#               'STARTFILTERS\n'
#               'ENDFILTERS\n\n')
#     anchors = sorted(anchors, key=lambda a: a.ref.start)
#     contentnt = []
#     contentcds = []
#     contentaa = []
#     #for winlen, *_, aaref, r in anchors:
#     for a in anchors:
#         #for seqid, (sim, i, j, _) in r.items():
#         for f in a:
#             # anchor2	C_CSFV_KC533775	-1	130	150	anchorsim	1.0
#             # anhcor100	D_BDV_NC_003679	-1	10	20	anchorxx
#             wlen = a.ref.len
#             i = f.start
#             j = f.seqstart
#             contentaa.append(
#                   #f'anchorsim\t{seqid}\t-1\t{i+1}\t{i+winlen}\tanchorsim\t{sim}\n'
#                  f'{a.ref.str[:5]} w{wlen}\t{f.id}\t-1\t{i+1}\t{i+wlen}\tanchor\n')
#             contentnt.append(
#                   #f'anchorsim\t{seqid}\t-1\t{j+1}\t{j+3*winlen}\tanchorsim\t{sim}\n'
#                  f'{a.ref.str[:5]} w{wlen}\t{f.id}\t-1\t{j+1}\t{j+3*wlen}\tanchor\n')
#             contentcds.append(
#                   #f'anchorsim\t{seqid}\t-1\t{j+1}\t{j+3*winlen}\tanchorsim\t{sim}\n'
#                  f'{a.ref.str[:5]} w{wlen}\t{f.id}\t-1\t{3*i+1}\t{3*i+3*wlen}\tanchor\n')

#     if fnameaa:
#         with open(fnameaa, 'w') as f:
#             f.write(header + ''.join(contentaa))
#     if fnament:
#         with open(fnament, 'w') as f:
#             f.write(header + ''.join(contentnt))
#     if fnamecds:
#         with open(fnamecds, 'w') as f:
#             f.write(header + ''.join(contentcds))


# def __write_jalview_features(anchors, fnameaa, fnament, fnamecds):
#     anchors = sorted(anchors, key=lambda a: a.ref.start)
#     contentnt = []
#     contentcds = []
#     contentaa = []
#     header = []
#     #for winlen, *_, aaref, r in anchors:
#     for k, a in enumerate(anchors):
#         #for seqid, (score, i, j, _) in r.items():
#         for j, f in enumerate(a):
#             # anchor2	C_CSFV_KC533775	-1	130	150	anchorsim	1.0
#             # anhcor100	D_BDV_NC_003679	-1	10	20	anchorxx
#             wlen = a.ref.len
#             i = f.start
#             j = f.seqstart
#             score_ = f.score / a.maxscore
#             c = to_hex(make_rgb_transparent(cols[k % len(cols)], 'white', score_)).strip('#')

#             al = f'anchor{k}_s{f.score}'
#             miss = f'misses:{len(a.missing)}'
#             header.append(
#                 f'{al}\t{c}\n'
#                 )
#             contentaa.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{i+1}\t{i+wlen}\t{al}\n')
#             contentnt.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{j+1}\t{j+3*wlen}\t{al}\n')
#             contentcds.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{3*i+1}\t{3*i+3*wlen}\t{al}\n')

#     header.extend(['STARTFILTERS\n', 'ENDFILTERS\n\n'])
#     header = ''.join(header)

#     if fnameaa:
#         with open(fnameaa, 'w') as f:
#             f.write(header + ''.join(contentaa))
#     if fnament:
#         with open(fnament, 'w') as f:
#             f.write(header + ''.join(contentnt))
#     if fnamecds:
#         with open(fnamecds, 'w') as f:
#             f.write(header + ''.join(contentcds))


# def __write_jalview_features(anchors, fnameaa, fnament, fnamecds):
#     anchors = sorted(anchors, key=lambda a: a.ref.start)
#     contentnt = []
#     contentcds = []
#     contentaa = []
#     header = []
#     #for winlen, *_, aaref, r in anchors:
#     for k, a in enumerate(anchors):
#         #for seqid, (score, i, j, _) in r.items():
#         for j, f in enumerate(a):
#             # anchor2	C_CSFV_KC533775	-1	130	150	anchorsim	1.0
#             # anhcor100	D_BDV_NC_003679	-1	10	20	anchorxx
#             wlen = a.ref.len
#             i = f.start
#             j = f.seqstart
#             score_ = f.score / a.maxscore
#             c = to_hex(make_rgb_transparent(cols[k % len(cols)], 'white', score_)).strip('#')

#             al = f'anchor{k}_s{f.score}'
#             miss = f'misses:{len(a.missing)}'
#             header.append(
#                 f'{al}\t{c}\n'
#                 )
#             contentaa.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{i+1}\t{i+wlen}\t{al}\n')
#             contentnt.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{j+1}\t{j+3*wlen}\t{al}\n')
#             contentcds.append(
#                  f'{a.ref.str[:5]} w{wlen} {miss}\t{f.id}\t-1\t{3*i+1}\t{3*i+3*wlen}\t{al}\n')

#     header.extend(['STARTFILTERS\n', 'ENDFILTERS\n\n'])
#     header = ''.join(header)

#     if fnameaa:
#         with open(fnameaa, 'w') as f:
#             f.write(header + ''.join(contentaa))
#     if fnament:
#         with open(fnament, 'w') as f:
#             f.write(header + ''.join(contentnt))
#     if fnamecds:
#         with open(fnamecds, 'w') as f:
#             f.write(header + ''.join(contentcds))



# (C) 2024, Tom Eulenfeld, MIT license
import contextlib
import io
import os
from pathlib import Path
from subprocess import check_output
import tempfile
from unittest.mock import patch

from sugar import read

from anchorna import cutout, read_anchors
from anchorna.cli import run_cmdline
from anchorna.io import load_json, load_selected_anchors, write_json


_IDS = (  # Representative sequences of pesti virus
    'NC_076029 NC_001461 KC853440 JX419398 M96687 KT951841 ON165517 KX577637 '
    'MW054939 AF526381 JQ799141 NC_076032 NC_039237 MH231152 KT875135 '
    'MH806436 HG426490 MK599227 JN380086 NC_038912 NC_002657 OM817567 '
    'GU233732 KP343640 KC533775 AY805221 AF099102 AY775178 MG387218 '
    'NC_003679 MF102261 NC_024018 NC_023176 NC_003678 NC_012812 NC_018713 '
    'NC_025677 NC_038964 NC_030653 MK216749 MH221025 MN099165 MN584738 '
    'NC_035432 NC_077026 MZ664274 NC_077023 NC_077024 ON024093 ON024108 '
    'NC_077001 NC_077000 NC_077015 OU592965 OM030319 OM030320').split()


def create_example_seqs_file():
    """
    Function to create the example sequences file in the tests/data folder
    """
    from sugar.web import Entrez
    seqs = Entrez().get_basket(_IDS)
    for seq in seqs:
        seq.fts = seq.fts.select('cds')
    seqs.write('data/pesti56.gff', archive='zip')


@contextlib.contextmanager
def _changedir(path):
    origin = Path().resolve()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)

@contextlib.contextmanager
def _changetmpdir():
    origin = Path().resolve()
    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            os.chdir(tmpdir)
            yield Path(tmpdir)
        finally:
            os.chdir(origin)

def check(cmd):
    args = cmd.split()
    assert args[0] == 'anchorna'
    with contextlib.redirect_stdout(io.StringIO()) as f:
        run_cmdline(args[1:])
    return f.getvalue()


def test_anchorna_workflow_subset():
    """
    Tests the full anchorna workflow with a subset of the example sequences
    """
    with _changetmpdir() as tmpdir:
        assert '' == check('anchorna create')
        fname_seqs = tmpdir / 'pesti_example.gff'
        assert '' == check('anchorna create --tutorial-subset')
        assert '' == check('anchorna go --no-pbar anchors.gff')
        assert 'A11' in check('anchorna print anchors.gff')
        assert 'F1' in check('anchorna print anchors.gff -v')
        out1 = check('anchorna combine anchors.gff|a5:a10|a8')
        out2 = check('anchorna combine anchors.gff|a5,a6,a7,a9')
        assert out1 == out2
        assert 'anchor0' in check('anchorna export anchors.gff')
        assert 'anchor0' in check('anchorna export -m seq anchors.gff')
        assert 'anchor0' in check('anchorna export -m aa anchors.gff')
        assert 'anchor0' in check('anchorna export -m cds anchors.gff')
        with patch('subprocess.run'):  # we do not want to actually start jalview here
            assert '' == check('anchorna view anchors.gff')
            assert '' == check('anchorna view anchors.gff --align a1')

        anchors = read_anchors('anchors.gff')
        assert len(anchors) == 15

        # test cutout
        assert '>' in check(f'anchorna cutout anchors.gff atg>+5 end-10 --fname {fname_seqs}')

        seqs = read(fname_seqs)
        seqs2 = cutout(seqs, anchors, 'start+10', 'a5-5')
        seqs3 = cutout(seqs, anchors, 'a5-5', '*>')
        seqs4 = cutout(seqs, anchors, '*>', 'end')
        assert str(seqs[0, 10:]) == str(seqs2[0] + seqs3[0] + seqs4[0])

        fname = tmpdir / 'pesti_test_cutout.sjson'
        assert '' == check(f'anchorna cutout anchors.gff a0> a2< -o {fname}')
        assert '' == check(f'anchorna go --fname {fname} --no-pbar anchors_cutout.gff')
        assert '' == check('anchorna combine anchors.gff||a1 anchors_cutout.gff -o anchors_combined.gff')
        assert read_anchors('anchors_combined.gff') == anchors

        fname = tmpdir / 'pesti_test_cutout2.sjson'
        assert '' == check(f'anchorna cutout anchors.gff a6> a10< -o {fname}')
        assert '' == check(f'anchorna go --fname {fname} --no-pbar anchors_cutout2.gff --maxshift=1000')
        assert '' == check('anchorna combine anchors.gff||a7:a10 anchors_cutout2.gff -o anchors_combined2.gff')
        assert read_anchors('anchors_combined2.gff') == load_selected_anchors('anchors.gff||a7')  # A7 is overlapping with A& and therefore not found

        # TODO: add test with missing flukes

        # check --no-remove option and --continue-with option
        assert '' == check('anchorna go --no-remove --no-pbar anchors2.gff')
        assert '' == check('anchorna go --continue-with anchors2.gff --no-pbar anchors3.gff')
        assert len(anchors) < len(read_anchors('anchors2.gff'))
        assert anchors == read_anchors('anchors3.gff')

        # test json
        json = tmpdir / 'anchors.json'
        write_json(anchors, json)
        assert load_json(json) == anchors


def test_anchorna_script_help():
    """
    Test that anchorna can be called on the command line
    """
    assert b'Find anchors' in check_output('anchorna -h'.split())
    assert b'go' in check_output('anchorna go -h'.split())



# (C) 2023, Tom Eulenfeld, MIT license

import logging
from statistics import median
from warnings import warn

from sugar import BioBasket
from sugar.data import submat
from tqdm import tqdm

from anchorna.util import _apply_mode, corrscore, Anchor, AnchorList, Fluke


log = logging.getLogger('anchorna')


def maxes(a, key=None, default=None):
    """
    Like max function, but returns list of *all* maximal values
    """
    if len(a) == 0:
        return default
    if key is None:
        key = lambda x: x
    kmax = key(a[0])
    max_list = []
    for s in a:
        k = key(s)
        if k > kmax:
            kmax = k
            max_list = [s]
        elif k == kmax:
            max_list.append(s)
    return max_list


def shift_and_find_best_word(seq, words, starti, w, sm, maxshift=None, maxshift_right=None):
    """
    Find position of the most similar word in a sequence compared to a set of words

    :return: tuple with similarity and index
    """
    seq = str(seq)
    if maxshift is None:
        maxshift = len(seq)
    if maxshift_right is None:
        maxshift_right = maxshift
    siminds = maxes([(max(corrscore(seq[i:i+w], word, sm=sm) for word in words), i)
                      for i in range(max(0, starti-maxshift_right),
                                     min(len(seq)-w, starti+maxshift+1))],
                     default=(0, None))
    if len(siminds) > 1:
        log.warning(f'multiple max values exist: {siminds=}')
    sim, ind = siminds[0]
    return sim, ind


def anchor_for_words(i, words, aas, w, refid, maxshift, thr_score, thr_quota_score, thr_quota,
                     scoring, method='default'):
    """
    Find the most similar not yet found word in a list of sequences and add it to words set

    If no good word is found anymore, return a single anchor.
    """
    assert thr_quota_score >= thr_score
    winlen = w
    aaref = [aa for aa in aas if aa.id == refid][0]
    assert aaref == aas[0]
    res = []
    nfails = 0
    for aa in aas:
        # if aa.id == seqrefid:
        #     continue
        maxshiftl = maxshift + max(0, len(aa) - len(aaref))
        maxshiftr = maxshift + max(0, len(aaref) - len(aa))
        score, start = shift_and_find_best_word(aa, words, i, winlen, submat(scoring), maxshiftl, maxshiftr)
        if start is None:
            raise ValueError('zero length sequence - that should not happen')
        if score < thr_quota_score:
            nfails += 1
            # short way out 50 % fail
            if nfails / len(aas) >= 0.5:
                return
        res.append((score, start, str(aa)[start:start+winlen], aa, score >= thr_score))
    if method != 'simple':
        for score, start, word, aa_, above_thres in sorted(res, reverse=True):
            if above_thres and word not in words:
                words.add(word)
                return 'continue'
    if nfails / len(aas) > 1 - thr_quota:
        return
    flukes = []
    for _, start, word, aa, above_thres in res:
        # score = median(corrscore(word, w) for w in words)  # this is done in anchor._calculate_scores()
        fluke=Fluke(seqid=aa.id, score=None, start=start, offset=aa.meta.get('offset'), stop=start+winlen, word=word, above_thres=above_thres)
        flukes.append(fluke)
    assert flukes[0].seqid == refid
    anchor = Anchor(flukes, refid=refid)
    anchor._calculate_scores()
    return anchor


def find_anchors_winlen(aas, options, indexrange=None, anchors=None, pbar=False):
    """
    Find multiple anchors in aa sequences for a specific word length
    """
    for i, aa in enumerate(aas):
        if aa.id == options.refid:
            break
    else:
        raise ValueError(f'No sequence with ref id {options.refid}')
    aas.insert(0, aas.pop(i))
    aaref = aas[0]
    if indexrange is None:
        indexrange = list(range(len(aaref)-options.w))
    if anchors is None:
        anchors = []
    if pbar:
        desc = '{:3d} anchors found, check candidates'
        pbar = tqdm(desc=desc.format(0), total=len(indexrange))
    for i in indexrange:
        words = {str(aaref)[i:i+options.w]}
        anchor = 'continue'
        while anchor == 'continue':
            anchor = anchor_for_words(i, words, aas, **options)
        if anchor is not None:
            anchors.append(anchor)
        if pbar:
            if pbar.update():
                pbar.set_description(desc.format(len(anchors)))
    anchors = sorted(anchors, key=lambda a: a.ref.start)
    return AnchorList(anchors, options=options)


def find_my_anchors(seqs, options, remove=True, continue_with=None, **kw):
    """
    Find and return anchors in CDS region of nucleotide sequences
    """
    if continue_with is None:
        all_offset = all('offset' in seq.meta for seq in seqs)
        all_cds = all('features' in seq.meta and seq.fts.get('cds') for seq in seqs)
        if all_offset:
            log.info('Found offsets in sequence file, translate full sequence')
            aas = seqs.translate(complete=True)
            for aa in aas:
                if '*' in str(aa):
                    log.error(f'Stop codon in aa sequence {aa.id}')
        elif all_cds:
            log.info('Found CDS features in sequence file, translate CDS')
            aas = seqs['cds'].translate()
            for aa in aas:
                aa.meta.offset = aa.fts.get('cds').loc.start
        else:
            log.warning('Did not found CDS annotation or offset for at least one sequence, do not translate')
            aas = seqs
        log.debug('result of translation are {}'.format(aas.tostr(h=0)))
        log.info(f'find anchors for word length {options.w}')
        anchors = find_anchors_winlen(aas, options, **kw)
        log.info(f'found {len(anchors)} anchors for word length {options.w}')
        anchors = anchors.merge_neighbor_anchors()
        log.info(f'merged into {len(anchors)} anchors')
    else:
        anchors = continue_with
    if remove:
        removed_anchors = anchors.remove_contradicting_anchors()
        removed_anchors.data = sorted(removed_anchors, key=lambda a: a.ref.start)
        log.info(f'After removal of contradicting anchors {len(anchors)} anchors left')
    else:
        removed_anchors = None
    anchors.data = sorted(anchors, key=lambda a: a.ref.start)
    assert all([f[0].seqid == options.refid for f in anchors])
    return anchors, removed_anchors


def _split_cutout_pos(pos, mode, seqs, anchors):
    """
    Split a string, .i.e. "A10>+5", into its three parts A, B, C
    """

    if (sign:='+') in pos or (sign:='-') in pos:
        pos, C = pos.split(sign)
        C = int(sign + C)
    else:
        C = 0
    if (B := pos[-1]) in '<>^':
        pos = pos[:-1]
        if pos in ('start', 'end'):
            raise ValueError('<>^ alignment characters not allowed for start, end')
    else:
        B = '^'
    if pos == 'start' and C < 0:
        raise ValueError('C<0 not allowed for start')
    if pos == 'end' and C > 0:
        raise ValueError('C>0 not allowed for end')
    if pos in ('atg', '*') and mode != 'seq':
        raise ValueError(f'{pos} only allowed in mode seq')
    A = pos
    if is_real_anchor := (A not in ('start', 'end', 'atg', '*')):
        A = anchors[int(A.removeprefix('a'))].d
        if len(ids := set(A.keys()) - set(seqs.keys())) > 0:
            warn(f'Some anchor ids {ids} not present in sequences')
    return A, B, C, is_real_anchor


def _transform_cutout_index(A, B, C, id_, seq, mode):
    """
    Transform a fluke and position given by A,B,C to and an integer index
    """
    if A == 'start':
        i1 = i2 = 0
    elif A == 'end':
        i1 = i2 = len(seq)
    elif A == 'atg':
        assert mode == 'seq'
        i1 = seq.fts.get('cds').loc.start
        i2 = i1 + 3
    elif A == '*':
        assert mode == 'seq'
        i2 = seq.fts.get('cds').loc.stop
        i1 = i2 - 3
    else:
        i1 = _apply_mode(A[id_].start, A[id_].offset, mode=mode)
        i2 = _apply_mode(A[id_].stop, A[id_].offset, mode=mode)
    i = i1 if B == '<' else i2 if B == '>' else (i1+i2)//2
    i += C
    return i


def cutout(seqs, anchors, pos1, pos2, mode='seq'):
    """
    Cutout subsequences from pos1 to pos2 (i.e. between two anchors)
    """
    seqs = seqs.d
    la, lb, lc, l_is_real_anchor = _split_cutout_pos(pos1.strip().lower(), mode, seqs, anchors)
    ra, rb, rc, r_is_real_anchor = _split_cutout_pos(pos2.strip().lower(), mode, seqs, anchors)
    seqs2 = BioBasket()
    for id_ in seqs:
        if (l_is_real_anchor and id_ not in la) or (r_is_real_anchor and id_ not in ra):
            # no fluke for this sequence
            continue
        i = _transform_cutout_index(la, lb, lc, id_, seqs[id_], mode)
        j = _transform_cutout_index(ra, rb, rc, id_, seqs[id_], mode)
        seq2 = seqs[id_][i:j]
        if mode == 'seq':
            seq2.meta.offset = i
            seq2.meta.pop('features', None)
        seqs2.append(seq2)
    return seqs2


def combine(lot_of_anchors):
    """
    Combine lists of anchors into a single AnchorList

    Deal with possibly different offset values.
    """
    anchors = set()
    offsets = {f.seqid: f.offset for anchor in lot_of_anchors[0] for f in anchor}
    for nans in lot_of_anchors:
        if len(set(nans) & anchors) > 0:
            ids = ', '.join(a.id for a in nans & anchors)
            raise ValueError(f'Anchors {ids} exits in multiple files')
        for anchor in nans:
            for f in anchor:
                doff = f.offset - offsets[f.seqid]
                f.offset = offsets[f.seqid]
                f.start += doff // 3
                f.stop += doff // 3
        anchors |= set(nans)
    return AnchorList(sorted(anchors, key=lambda a: a.ref.start))
